import numpy as np

from .vikor import VIKOR
from ..additions import rank_preferences


class VIKOR_SMAA():
    def __init__(self, normalization_method = None, v = 0.5):
        """Create the VIKOR method object.

        Parameters
        -----------

            normalization_method : function
                VIKOR does not use normalization by default, thus `normalization_method` is set to None by default.
                However, you can choose method for normalization of decision matrix chosen `normalization_method` from `normalizations`.
                It is used in a way `normalization_method(X, types)` where `X` is a decision matrix
                and `types` is a vector with criteria types where 1 means profit and -1 means cost.
            v : float
                parameter that is the weight of strategy of the majority of criteria (the maximum group utility)
        """
        self.v = v
        self.normalization_method = normalization_method

    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.
        
        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights : ndarray
                Matrix with i vectors in rows of n weights in columns. i means number of
                iterations of SMAA
            types : ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.
        
        Returns
        --------
            ndrarray, ndarray, ndarray
                Matrix with acceptability indexes values for each alternative in rows in relation to each rank in columns,
                Matrix with central weight vectors for each alternative in rows
                Matrix with final ranking of alternatives
        
        Examples
        ---------
        >>> vikor_smaa = VIKOR_SMAA(normalization_method = minmax_normalization)
        >>> rank_acceptability_index, central_weight_vector, rank_scores = vikor_smaa(matrix, weights, types)
        """

        return VIKOR_SMAA._vikor_smaa(self, matrix, weights, types, self.normalization_method, self.v)


    # function to generate multiple weight vectors
    # returns matrix with n weights in columns and number of vectors in rows equal to iterations number
    def _generate_weights(self, n, iterations):
        """
        Function to generate multiple weight vectors

        Parameters
        -----------
            n : int
                Number of criteria
            iterations : int
                Number of weight vector to generate

        Returns
        ----------
            ndarray
                Matrix containing in rows vectors with weights for n criteria

        """
        weight_vectors = np.zeros((iterations, n))
        # n weight generation - when no preference information available
        # generate n - 1 uniform distributed weights within the range [0, 1]
        for i in range(iterations):
            w = np.random.uniform(0, 1, n)

            # sort weights into ascending order (q[1], ..., q[n-1])
            ind = np.argsort(w)
            w = w[ind]

            # insert 0 as the first q[0] and 1 as the last (q[n]) numbers
            w = np.insert(w, 0, 0)
            w = np.insert(w, len(w), 1)

            # the weights are obtained as intervals between consecutive numbers (w[j] = q[j] - q[j-1])
            weights = [w[i] - w[i - 1] for i in range(1, n + 1)]
            weights = np.array(weights)

            # scale the generated weights so that their sum is 1
            new_weights = weights / np.sum(weights)
            weight_vectors[i, :] = new_weights
        return weight_vectors


    @staticmethod
    def _vikor_smaa(self, matrix, weights, types, normalization_method, v):
        m, n = matrix.shape

        # Central weight vector for each alternative
        central_weight_vector = np.zeros((m, n))
        
        # Rank acceptability index of each place for each alternative
        rank_acceptability_index = np.zeros((m, m))

        # Ranks
        rank_score = np.zeros(m)

        vikor = VIKOR()
    
        # Calculate alternatives preference function values with VIKOR method
        pref = vikor(matrix, weights, types)

        # Calculate rankings based on preference values
        rank = np.zeros((pref.shape))
        for i in range(pref.shape[1]):
            rank[:, i] = rank_preferences(pref[:, i], reverse = False)

            # add value for the rank acceptability index for each alternative considering rank and rank score
            # iteration by each alternative
            rr = rank[:, i]
            for k, r in enumerate(rr):
                rank_acceptability_index[k, int(r - 1)] += 1
                # rank score
                # calculate how many alternatives have worst preference values than k-th alternative
                # Note: in VIKOR better alternatives have lower preference values
                better_ranks = rr[rr > rr[k]]
                # add to k-th index value 1 for each alternative that is worse than k-th alternative
                rank_score[k] += len(better_ranks)
            
            # add central weights for the best scored alternative
            ind_min = np.argmin(rr)
            central_weight_vector[ind_min, :] += weights[i, :]

        #
        # end of loop for i iterations
        # Calculate the rank acceptability index
        rank_acceptability_index = rank_acceptability_index / pref.shape[1]

        # Calculate central the weights vectors
        central_weight_vector = central_weight_vector / pref.shape[1]
        for i in range(m):
            if np.sum(central_weight_vector[i, :]):
                central_weight_vector[i, :] = central_weight_vector[i, :] / np.sum(central_weight_vector[i, :])

        # Calculate rank scores
        rank_score = rank_score / pref.shape[1]
        rank_scores = rank_preferences(rank_score, reverse = True)

        return rank_acceptability_index, central_weight_vector, rank_scores


import copy
import sys
import itertools

import numpy as np
from .correlations import pearson_coeff
from .normalizations import sum_normalization, minmax_normalization



# Equal weighting
def equal_weighting(matrix):
    """
    Calculate criteria weights using objective Equal weighting method.

    Parameters
    -----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.

    Returns
    --------
        ndarray
            Vector of criteria weights.

    Examples
    ----------
    >>> weights = equal_weighting(matrix)
    """
    N = np.shape(matrix)[1]
    w = np.ones(N) / N
    return w


# Entropy weighting
def entropy_weighting(matrix):
    """
    Calculate criteria weights using objective Entropy weighting method.

    Parameters
    -----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.

    Returns
    --------
        ndarray
            Vector of criteria weights.

    Examples
    ----------
    >>> weights = entropy_weighting(matrix)
    """
    # normalize the decision matrix with `sum_normalization` method from `normalizations` as for profit criteria
    types = np.ones(np.shape(matrix)[1])
    pij = sum_normalization(matrix, types)
    # Transform negative values in decision matrix `matrix` to positive values
    pij = np.abs(pij)
    m, n = np.shape(pij)
    H = np.zeros((m, n))

    # Calculate entropy
    for j, i in itertools.product(range(n), range(m)):
        if pij[i, j]:
            H[i, j] = pij[i, j] * np.log(pij[i, j])

    h = np.sum(H, axis = 0) * (-1 * ((np.log(m)) ** (-1)))

    # Calculate degree of diversification
    d = 1 - h

    # Set w as the degree of importance of each criterion
    w = d / (np.sum(d))
    return w


# Standard Deviation weighting
def std_weighting(matrix):
    """
    Calculate criteria weights using objective Standard deviation weighting method.

    Parameters
    -----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
            
    Returns
    --------
        ndarray
            Vector of criteria weights.

    Examples
    ----------
    >>> weights = std_weighting(matrix)
    """
    
    # Calculate the standard deviation of each criterion in decision matrix
    stdv = np.sqrt((np.sum(np.square(matrix - np.mean(matrix, axis = 0)), axis = 0)) / (matrix.shape[0]))
    # Calculate criteria weights by dividing the standard deviations by their sum
    w = stdv / np.sum(stdv)
    return w


# CRITIC weighting
def critic_weighting(matrix):
    """
    Calculate criteria weights using objective CRITIC weighting method.

    Parameters
    -----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
            
    Returns
    --------
        ndarray
            Vector of criteria weights.

    Examples
    ----------
    >>> weights = critic_weighting(matrix)
    """
    # Normalize the decision matrix using Minimum-Maximum normalization `minmax_normalization` from `normalizations` as for profit criteria
    types = np.ones(np.shape(matrix)[1])
    x_norm = minmax_normalization(matrix, types)
    # Calculate the standard deviation
    std = np.std(x_norm, axis = 0)
    n = np.shape(x_norm)[1]
    # Calculate correlation coefficients of all pairs of columns of normalized decision matrix
    correlations = np.zeros((n, n))
    for i, j in itertools.product(range(n), range(n)):
        correlations[i, j] = pearson_coeff(x_norm[:, i], x_norm[:, j])

    # Calculate the difference between 1 and calculated correlations
    difference = 1 - correlations
    # Multiply the difference by the standard deviation
    C = std * np.sum(difference, axis = 0)
    # Calculate the weights by dividing vector with `C` by their sum
    w = C / np.sum(C)
    return w


# Gini coefficient-based weighting
def gini_weighting(matrix):
    """
    Calculate criteria weights using objective Gini coefficient-based weighting method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
            
    Returns
    --------
        ndarray
            Vector of criteria weights.

    Examples
    ---------
    >>> weights = gini_weighting(matrix)
    """
    m, n = np.shape(matrix)
    G = np.zeros(n)
    # Calculate the Gini coefficient for decision matrix `matrix`
    # iteration over criteria j = 1, 2, ..., n
    for j in range(n):
        Yi = 0
        # iteration over alternatives i = 1, 2, ..., m
        if np.mean(matrix[:, j]):
            for i in range(m):
                Yi += np.sum(np.abs(matrix[i, j] - matrix[:, j]) / (2 * m**2 * (np.sum(matrix[:, j]) / m)))
        else:
            for i in range(m):
                Yi += np.sum(np.abs(matrix[i, j] - matrix[:, j]) / (m**2 - m))

        G[j] = Yi
    # calculate and return the criteria weights by dividing the vector of Gini coefficients by their sum
    w = G / np.sum(G)
    return w


# MEREC weighting
def merec_weighting(matrix, types):
    """
    Calculate criteria weights using objective MEREC weighting method.

    Parameters
    -----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
        types : ndarray
            Vector with criteria types.
            
    Returns
    --------
        ndarray
            Vector of criteria weights.

    Examples
    ---------
    >>> weights = merec_weighting(matrix, types)
    """
    X = copy.deepcopy(matrix)
    m, n = X.shape
    # Transform negative values in decision matrix X to positive values
    X = np.abs(X)
    # Normalize the decision matrix X with linear normalization method
    norm_matrix = np.zeros(X.shape)
    norm_matrix[:, types == 1] = np.min(X[:, types == 1], axis = 0) / X[:, types == 1]
    norm_matrix[:, types == -1] = X[:, types == -1] / np.max(X[:, types == -1], axis = 0)
    
    # Calculate the overall performance of the values in normalized matrix using a logarithmic measure with equal criteria weights
    S = np.log(1 + ((1 / n) * np.sum(np.abs(np.log(norm_matrix)), axis = 1)))

    # Calculate the performance of the alternatives by removing each criterion using the logarithmic measure
    Sp = np.zeros(X.shape)

    for j in range(n):
        norm_mat = np.delete(norm_matrix, j, axis = 1)
        Sp[:, j] = np.log(1 + ((1 / n) * np.sum(np.abs(np.log(norm_mat)), axis = 1)))

    # Calculate the summation of absolute deviations
    E = np.sum(np.abs(Sp - S.reshape(-1, 1)), axis = 0)

    # Calculate the final weights of the criteria
    w = E / np.sum(E)
    return w


# Statistical Variance weighting
def stat_var_weighting(matrix):
    """
    Calculate criteria weights using objective Statistical variance weighting method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
            
    Returns
    -------
        ndarray
            Vector of criteria weights.

    Examples
    ---------
    >>> weights = stat_var_weighting(matrix)
    """

    # Normalize the decision matrix `matrix` with `minmax_normalization` method from normalizations
    types = np.ones(np.shape(matrix)[1])
    xn = minmax_normalization(matrix, types)
    # Calculate the statistical variance for each criterion
    v = np.mean(np.square(xn - np.mean(xn, axis = 0)), axis = 0)
    # Calculate the final weights of the criteria
    w = v / np.sum(v)
    return w


# CILOS weighting
def cilos_weighting(matrix, types):
    """
    Calculate criteria weights using objective CILOS weighting method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
        types : ndarray
            Vector with criteria types.
            
    Returns
    -------
        ndarray
            Vector of criteria weights.

    Examples
    >>> weights = cilos_weighting(matrix, types)
    """
    xr = copy.deepcopy(matrix)
    # Convert negative criteria to positive criteria
    xr[:, types == -1] = np.min(matrix[:, types == -1], axis = 0) / matrix[:, types == -1]
    # Normalize the decision matrix `xr` using the sum normalization method
    xn = xr / np.sum(xr, axis = 0)
    
    # Calculate the square matrix
    A = xn[np.argmax(xn, axis = 0), :]
    
    # Calculate relative impact loss matrix
    pij = np.zeros((matrix.shape[1], matrix.shape[1]))
    for j, i in itertools.product(range(matrix.shape[1]), range(matrix.shape[1])):
        pij[i, j] = (A[j, j] - A[i, j]) / A[j, j]

    # Determine the weight system matrix
    F = np.diag(-np.sum(pij - np.diag(np.diag(pij)), axis = 0)) + pij
    # Calculate the criterion impact loss weight
    # The criteria weights q are determined from the formulated homogeneous linear system of equations
    # AA is the vector near 0
    AA = np.zeros(F.shape[0])
    # To determine the value of A we assume that the first element of A is close to 0 while others are zeros
    AA[0] = sys.float_info.epsilon
    # Solve the system equation
    q = np.linalg.inv(F).dot(AA)
    # Calculate and return the final weights of the criteria
    return q / np.sum(q)


# IDOCRIW weighting
def idocriw_weighting(matrix, types):
    """
    Calculate criteria weights using objective IDOCRIW weighting method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
        types : ndarray
            Vector with criteria types.
            
    Returns
    -------
        ndarray
            Vector of criteria weights.

    Examples
    ---------
    >>> weights = idocriw_weighting(matrix, types)
    """
    # Calculate the Entropy weights
    q = entropy_weighting(matrix)
    # Calculate the CILOS weights
    w = cilos_weighting(matrix, types)
    # Aggregate the weight value of the attributes considering Entropy and CILOS weights
    weights = (q * w) / np.sum(q * w)
    return weights


# Angle weighting
def angle_weighting(matrix, types):
    """
    Calculate criteria weights using objective Angle weighting method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
        types : ndarray
            Vector with criteria types.
            
    Returns
    -------
        ndarray
            Vector of criteria weights.

    Examples
    ---------
    >>> weights = angle_weighting(matrix, types)
    """
    m, n = matrix.shape
    # Normalize the decision matrix X using sum_normalization method from normalizations
    X = sum_normalization(matrix, types)
    # Calculate elements of additional column (the reference attribute) which are equal to 1 / m
    B = np.ones(m) * (1 / m)
    # Calculate the angle between attraibutes in decision matrix X and the reference attribute
    u = np.arccos(np.sum(X / m, axis = 0) / (np.sqrt(np.sum(X ** 2, axis = 0)) * np.sqrt(np.sum(B ** 2))))
    # Calculate the final angle weights for each criterion
    w = u / np.sum(u)
    return w


# Coeffcient of variation weighting
def coeff_var_weighting(matrix):
    """
    Calculate criteria weights using objective Coefficient of variation weighting method.
    
    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria.
            
    Returns
    -------
        ndarray
            Vector of criteria weights.

    Examples
    ---------
    >>> weights = coeff_var_weighting(matrix)
    """
    m, n = matrix.shape
    # Normalize the decision matrix `matrix` with `sum_normalization` method from `normalizations`
    types = np.ones(n)
    B = sum_normalization(matrix, types)

    # Calculate the standard deviation of each column
    Bm = np.sum(B, axis = 0) / m
    std = np.sqrt(np.sum(((B - Bm)**2), axis = 0) / (m - 1))

    # Calculate the Coefficient of Variation for each criterion
    ej = std / Bm
    # Calculate the weights for each criterion
    w = ej / np.sum(ej)
    return w


import numpy as np

from .mcda_method import MCDA_method


class VIKOR(MCDA_method):
    def __init__(self, normalization_method = None, v = 0.5):
        """Create the VIKOR method object.

        Parameters
        -----------

            normalization_method : function
                VIKOR does not use normalization by default, thus `normalization_method` is set to None by default.
                However, you can choose method for normalization of decision matrix chosen `normalization_method` from `normalizations`.
                It is used in a way `normalization_method(X, types)` where `X` is a decision matrix
                and `types` is a vector with criteria types where 1 means profit and -1 means cost.
            v : float
                parameter that is the weight of strategy of the majority of criteria (the maximum group utility)
        """
        self.v = v
        self.normalization_method = normalization_method

    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.
        
        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Matrix containing vectors with criteria weights in subsequent rows. 
                Sum of weights in each vector must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.
        
        Returns
        --------
            ndrarray
                Matrix with vectors containing preference values of each alternative. 
                The best alternative has the lowest preference value. Vectors are placed
                in subsequent columns of matrix.
        
        Examples
        ---------
        >>> vikor = VIKOR(normalization_method = minmax_normalization)
        >>> pref = vikor(matrix, weights, types)
        >>> rank = np.zeros((pref.shape))
        >>> for i in range(pref.shape[1]):
        >>>     rank[:, i] = rank_preferences(pref[:, i], reverse = False)
        """
        VIKOR._verify_input_data(matrix, weights, types)
        return VIKOR._vikor(matrix, weights, types, self.normalization_method, self.v)

    @staticmethod
    def _vikor(matrix, weights, types, normalization_method, v):
        # if weights are vector with one dimension
        if len(weights.shape) == 1:
            weights = weights.reshape(1, -1)
        # array for collecting vectors with preference values in subsequent columns
        pref = np.zeros((matrix.shape[0], weights.shape[0]))
        for el, w in enumerate(weights):
            # Without applying a special normalization method
            if normalization_method == None:

                # Determine the best `fstar` and the worst `fmin` values of all criterion function
                maximums_matrix = np.amax(matrix, axis = 0)
                minimums_matrix = np.amin(matrix, axis = 0)

                fstar = np.zeros(matrix.shape[1])
                fmin = np.zeros(matrix.shape[1])

                # for profit criteria (`types` == 1) and for cost criteria (`types` == -1)
                fstar[types == 1] = maximums_matrix[types == 1]
                fstar[types == -1] = minimums_matrix[types == -1]
                fmin[types == 1] = minimums_matrix[types == 1]
                fmin[types == -1] = maximums_matrix[types == -1]

                weighted_matrix = w * ((fstar - matrix) / (fstar - fmin))
            else:
                # With applying the special normalization method
                norm_matrix = normalization_method(matrix, types)
                fstar = np.amax(norm_matrix, axis = 0)
                fmin = np.amin(norm_matrix, axis = 0)
                weighted_matrix = w * ((fstar - norm_matrix) / (fstar - fmin))

            # Calculate the `S` and `R` values
            S = np.sum(weighted_matrix, axis = 1)
            R = np.amax(weighted_matrix, axis = 1)
            # Calculate the Q values
            Sstar = np.min(S)
            Smin = np.max(S)
            Rstar = np.min(R)
            Rmin = np.max(R)
            Q = v * (S - Sstar) / (Smin - Sstar) + (1 - v) * (R - Rstar) / (Rmin - Rstar)
            # save vector with preference values in array `pref`
            pref[:, el] = Q
        return pref


import numpy as np

# reverse = True: descending order ( for example TOPSIS, CODAS), False: ascending order (for example VIKOR, SPOTIS)
def rank_preferences(pref, reverse = True):
    """
    Rank alternatives according to MCDA preference function values. If more than one alternative
    have the same preference function value, they will be given the same rank value (tie).

    Parameters
    ------------
        pref : ndarray
            Vector with MCDA preference function values for alternatives
        reverse : bool
            The boolean variable is True for MCDA methods that rank alternatives in descending
            order (for example, TOPSIS, CODAS) and False for MCDA methods that rank alternatives in ascending
            order (for example, VIKOR, SPOTIS)
    
    Returns
    ---------
        ndarray
            Vector with alternatives ranking. Alternative with 1 value is the best and has the first position in the ranking.
    
    Examples
    ----------
    >>> rank = rank_preferences(pref, reverse = True)
    """

    # Create an array ndarray for the ranking values of alternatives
    rank = np.zeros(len(pref))
    # Generate sorted vector with MCDA preference function values
    # sorting order is determined by variable `reverse` and depends on MCDA method
    sorted_pref = sorted(pref, reverse = reverse)
    # position of the best alternative in ranking is denoted by 1, so assign 1 to `pos`
    pos = 1
    for i in range(len(sorted_pref) - 1):
        # find index in vector with preference values `pref` equal to preference value in sorted vector `sorted_pref`
        ind = np.where(sorted_pref[i] == pref)[0]
        # assign rank denoted by `pos` to index `ind` in rank vector
        rank[ind] = pos
        # if the next preference value in sorted vector is higher than actual, increase `pos` which denotes rank
        # if the next preference value is equal to actual, `pos` will be unchanged
        if sorted_pref[i] != sorted_pref[i + 1]:
            pos += 1
    # find index with the last preference value
    ind = np.where(sorted_pref[i + 1] == pref)[0]
    # assign the last place `pos` to alternative with the last preference value
    rank[ind] = pos
    return rank.astype(int)